# -*- coding: utf-8 -*-
# Copyright 2025 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unittests for the CIF reader.
"""

import vermouth
from vermouth import CIFInput
from vermouth.tests.datafiles import CIF_PROTEIN

def test_CIFInput():

    system = vermouth.System()
    CIFInput(filename=CIF_PROTEIN, exclude='HOH', ignh=False, modelidx=1).run_system(system)

    assert len(system.molecules) == 1
